var canvas = document.getElementById("cvGame");
var ctx = canvas.getContext("2d");
var runway, bird;
var Sprite = function(imgSrc) {
    this.x = 0;
    this.y = 0;
    this.moveX = 0;
    this.moveY = 0;
    this.img = new Image();
    this.img.src = imgSrc;
    this.rotation = 0;
    this.frames = 1;
    this.frameWidth = 0;
    this.frame = 0;
    this.continuous = false;
    this.visible = false;
};
Sprite.prototype.animate = function() {
    if(this.visible){
        ctx.save();
        xAdj = (this.frameWidth > 0 ? this.frameWidth/2 : this.img.width/2);
        yAdj = this.img.height / 2;
        ctx.translate(this.x + xAdj, this.y + yAdj);
        ctx.rotate(this.rotation);
        if(this.frames > 1){
            if(game.state === game.states.start || game.state === game.states.running){
                this.frame += 1;
                this.frame %= this.frames;
            } else {
                this.frame = 0; // stop animation at end
            }
            let clipx = this.frame * this.frameWidth;
            ctx.drawImage(this.img, clipx, 0, this.frameWidth, this.img.height, -xAdj, -yAdj, this.frameWidth, this.img.height);
        } else {
            ctx.drawImage(this.img, -xAdj, -yAdj);
        }
        ctx.restore();
     }
     if(game.state === game.states.running || game.state === game.states.over){
        this.x += this.moveX;
        this.y += this.moveY;
        if(this.continuous && this.x < (canvas.width - this.img.width)){
            this.x = 0;
        }
    }
};
Sprite.prototype.repeatFrame = function(nctx) {
    nctx.save();
    xAdj = (this.frameWidth > 0 ? this.frameWidth/2 : this.img.width/2);
    yAdj = this.img.height / 2;
    nctx.translate(this.x + xAdj, this.y + yAdj);
    nctx.rotate(this.rotation);
    if(this.frames > 1){
        let clipx = this.frame * this.frameWidth;
        nctx.drawImage(this.img, clipx, 0, this.frameWidth, this.img.height, -xAdj, -yAdj, this.frameWidth, this.img.height);
    } else {
        ctx.drawImage(this.img, -xAdj, -yAdj);
    }
     nctx.restore();
};
var Rect = function(sprite, applyRotation){
    this.x = sprite.x;
    this.y = sprite.y;
    this.width = (sprite.frames > 1) ? sprite.frameWidth : sprite.img.width;
    this.height = sprite.img.height;
    if(applyRotation && sprite.rotation != 0){
        let rot = (sprite.rotation < 0) ? 2 * Math.PI - sprite.rotation : sprite.rotation;
        let w = Math.ceil(Math.sin(rot) * this.height + Math.cos(rot) * this.width);
        let h = Math.ceil(Math.sin(rot) * this.width + Math.cos(rot) * this.height);
        this.x -= Math.ceil((w - this.width)/2);
        this.y -= Math.ceil((h - this.height) / 2);
        this.width = w;
        this.height = h;
    }
};
function intersect(rectA, rectB){
    // return Math.max(rectA.x, rectB.x) < Math.min(rectA.x + rectA.width, rectB.x + rectB.width) &&
    //     Math.max(rectA.y, rectB.y) < Math.min(rectA.y + rectA.height, rectB.y + rectB.height);
    let rA = rectA.x + rectA.width, rB = rectB.x + rectB.width;//right x
    let bA = rectA.y + rectA.height, bB = rectB.y + rectB.height;//btm y
    let ix = Math.max(rectA.x, rectB.x), iy =  Math.max(rectA.y, rectB.y); //?? min
    let rect = {
        x: ix,
        y: iy,
        width: Math.min(rA, rB) - ix,
        height: Math.min(bA, bB) - iy
    };
    if(rect.x < (rect.x + rect.width) && rect.y < (rect.y + rect.height)){
        return rect;
    }
    return false;
}
var game = {
    states: Object.freeze({"start": 0, "running":1, "over": 2,"end" : 3}),
    speed: 50,
    move: -2,
    runwayHeight: 25,
    state: null,
    tmr: null,
    width: 320,
    height: 480,
    pipeHeight: 392,
    pipes: [],
    jump: -10,
    maxGravity: 10, // terminal velocity
    gravity: 1,
    score: 0,
    highScore: 0
};
var collision = {
    methods: Object.freeze({"rectangle": 0, "pixelcheck": 2}),
    canvas: document.createElement('canvas'),
    ctx: null,
    method: 2
};
function addPipePair(x, gapTop, gapHeight) {
    let topPipe = new Sprite("images/pipet.png");
    topPipe.x = x;
    topPipe.y = gapTop - game.pipeHeight;
    topPipe.moveX = game.move;
    game.pipes.push(topPipe);
    let bottomPipe = new Sprite("images/pipeb.png");
    bottomPipe.x = x;
    bottomPipe.y = gapTop + gapHeight;
    bottomPipe.moveX = game.move;
    game.pipes.push(bottomPipe);
}
function createPipes(){
    addPipePair(500, 100, 145);
    addPipePair(800, 50, 145);
    addPipePair(1000, 250, 140);
    addPipePair(1200, 150, 125);
    addPipePair(1600, 100, 125);
    addPipePair(1800, 150, 120);
    addPipePair(2000, 200, 120);
    addPipePair(2200, 250, 120);
    addPipePair(2400, 30, 105);
    addPipePair(2700, 300, 105);
    addPipePair(3000, 100, 85);
    addPipePair(3300, 250, 85);
    addPipePair(3600, 50, 65);
    // add finish bar here to keep things tidy
    let finish = new Sprite("images/finish.png");
    finish.x = 3900;
    finish.moveX = game.move;
    game.pipes.push(finish); // as finish must move with pipes.
}
function drawPipes(){
    for(let p = 0; p < game.pipes.length; p++) {
        //only draw pipes we can see but advance them all
        if(game.pipes[p].x > -(game.pipes[p].img.width) && game.pipes[p].x < canvas.width){
            game.pipes[p].visible = true;
        } else {
            game.pipes[p].visible = false;
        }
        game.pipes[p].animate();
    }
}

function initialise() {
    canvas.width = game.width;
    canvas.height = game.height;
    runway = new Sprite("images/runway.png");
    runway.x = 0;
    runway.y = canvas.height - game.runwayHeight;
    runway.moveX = game.move;
    runway.continuous = true;
    runway.visible = true;
    bird = new Sprite("images/flapping.png");
    bird.x = Math.floor(canvas.width / 3);
    bird.y = canvas.height / 2;
    bird.frames = 4;
    bird.frameWidth = 57;
    bird.visible = true;
    bird.angleUp = -15*Math.PI/180;
    bird.angleDip = 5*Math.PI/180;
    bird.maxDip = 70*Math.PI/180;
 
    createPipes();
    game.state = game.states.start;
    game.tmr = setInterval(drawFrame, game.speed);
		window.addEventListener("keydown", playerControl);
    window.addEventListener("pointerdown", playerControl); }
function playerControl(e) {
    switch (game.state){
       case game.states.start:
           game.state = game.states.running;
           break;
       case game.states.running:
           if(bird.y > game.jump){
            bird.moveY = game.jump;
           }
            break;
       case game.states.end:
           break;
   }
}
function drawFrame() {
    ctx.clearRect(0, 0, canvas.width, canvas.height);
    drawPipes();
    bird.animate();
    runway.animate();
    if(game.state == game.states.running) {
        birdStrike(); // did bird hit pipe ?
    }
    setBird();
    if(game.state === game.states.start){
        showPlay("Click, touch or key to start");
    }
}
function setBird() {
    if(bird.y > (canvas.height - (bird.img.height + game.runwayHeight))){
        game.state = game.states.end; // game ends here
        clearInterval(game.tmr); // stop the action
        return;
    }
    if(game.state === game.states.running || game.state === game.states.over){
        if(bird.moveY < game.maxGravity){
            bird.moveY += game.gravity;
        }
    }
    switch(true) {
        case bird.moveY <= -7:
            bird.rotation = bird.angleUp;
            break;
        case bird.moveY <=-4:
            bird.rotation = bird.angleUp + bird.angleDip;
            break;
        case bird.moveY <= 0:
            bird.rotation = 0;
            break;
        default:
            bird.rotation += (bird.rotation < bird.maxDip) ? bird.angleDip : 0;
            break;
    }
}
function birdStrike(){
    let bRect = new Rect(bird, true);
    for(let p = 0; p < game.pipes.length; p++) {
        if(game.pipes[p].visible){
            let pRect = new Rect(game.pipes[p]);
            let strike = intersect(bRect, pRect);
            if(strike && collision.method === collision.methods.pixelcheck){
                strike = checkPixels(strike);
            }
            if(strike) {
                game.state = game.states.over;
                return;
           }
        }
    }
}
function checkPixels(rect){
    if(!collision.ctx){
       collision.canvas.width = canvas.width;
       collision.canvas.height = canvas.height;
       collision.ctx = collision.canvas.getContext("2d");
   }
   collision.ctx.clearRect(0, 0, canvas.width, canvas.height);
   bird.repeatFrame(collision.ctx);
   let pixels = collision.ctx.getImageData(rect.x, rect.y, rect.width, rect.height);
   let pixelData = pixels.data;
   for(let p = 3; p < pixelData.length; p+=4){
       if(pixelData[p]!==0){
           return true;
       }
   }
   return false;
}
function showPlay(txt) {
    ctx.font= "24px Arial";
    ctx.fillStyle= "blue";
    ctx.textAlign="center";
    ctx.fillText(txt, canvas.width / 2, canvas.height / 2 + 60);  
}

initialise();